/**
 * Copyright 2023 Velocity Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const { nanoid } = require('nanoid/non-secure');
const { Client } = require('undici');
const CacheableLookup = require('cacheable-lookup');
const pkg = require('../package.json');

const USER_AGENT_HEADER = `${pkg.name}/${pkg.version}`;

const initHttpClient = (options) => {
  const { origin, rootPath, clientOptions, traceIdHeader, customHeaders } =
    parseOptions(options);
  const client = new Client(origin, clientOptions);

  const doRequest = async (url, reqOptions, method, { traceId, log }) => {
    const reqId = nanoid();
    const reqHeaders = {
      'user-agent': USER_AGENT_HEADER,
      [traceIdHeader]: traceId,
      ...customHeaders,
    };
    log.info({ url, reqId, reqHeaders }, 'HttpClient request');
    let path = rootPath;
    if (url != null) {
      // eslint-disable-next-line better-mutation/no-mutation
      path += `/${url}`;
    }
    if (reqOptions?.searchParams != null) {
      // eslint-disable-next-line better-mutation/no-mutation
      path += `?${reqOptions.searchParams}`;
    }
    const httpResponse = await client.request({
      path,
      method,
      headers: reqHeaders,
      throwOnError: true,
    });
    const { statusCode, headers: resHeaders, body } = httpResponse;
    return {
      statusCode,
      resHeaders,
      json: async () => {
        const bodyJson = await body.json();
        log.info(
          { url, reqId, statusCode, resHeaders, bodyJson },
          'HttpClient response'
        );
        return bodyJson;
      },
      text: async () => {
        const bodyText = await body.text();
        log.info(
          { url, reqId, statusCode, resHeaders, bodyText },
          'HttpClient response'
        );
        return bodyText;
      },
    };
  };

  return (context) => ({
    get: (url, reqOptions) =>
      doRequest(url, reqOptions, HTTP_VERBS.GET, context),
    responseType: 'promise',
  });
};

const parseOptions = (options) => {
  const url = new URL(options.prefixUrl);

  const clientOptions = {
    connect: {
      rejectUnauthorized: options.rejectUnauthorized ?? true,
      lookup: new CacheableLookup().lookup,
    },
  };
  if (options.requestTimeout != null) {
    clientOptions.bodyTimeout = options.requestTimeout;
  }

  return {
    origin: `${url.protocol}//${url.host}`,
    rootPath:
      url.pathname.at(-1) === '/' ? url.pathname.slice(0, -1) : url.pathname,
    clientOptions,
    traceIdHeader: options.traceIdHeader ?? 'TRACE_ID',
    customHeaders: options.customHeaders ?? {},
  };
};

const HTTP_VERBS = {
  GET: 'GET',
};

module.exports = { initHttpClient, parseOptions };
